from collections.abc import Iterable
from pathlib import Path


def get_sys_files(path: str | Path) -> Iterable[Path]:
    path = Path(path)
    if path.suffix == ".sys":
        return [path]
    return path.glob("**/*.sys")


def read_sys_file(path: str | Path) -> str:
    with Path(path).open() as f:
        return f.read()



import pyparsing as pp

from . import tokens as T

boolean_op = T.AND | T.OR | T.NOT | T.XOR
boolean_op.set_name("bool_op")

infix_op = pp.oneOf("+ - * / = <= < >= > <>") | boolean_op | T.DIV | T.MOD



from dataclasses import dataclass

from .config import CONFIG


@dataclass
class WhiteSpaceError:
    lineno: int
    col: int = 0
    error: str = ""
    suggestion: str = ""

    def __str__(self) -> str:
        return f"{self.lineno}:{self.col} {self.error}. => {self.suggestion}"


@dataclass
class TooManyEmptyLinesError(WhiteSpaceError):
    def __post_init__(self) -> None:
        self.error = "Too many empty lines"
        self.suggestion = f"Use up to {CONFIG.max_empty_lines} empty lines"


@dataclass
class TrailingSpaceError(WhiteSpaceError):
    def __post_init__(self) -> None:
        self.error = "Line ends with space"
        self.suggestion = "Remove trailing spaces"


@dataclass
class NoTrailingLineError(WhiteSpaceError):
    def __post_init__(self) -> None:
        self.error = "Line does not end with a new line character"
        self.suggestion = "Add a new line character at the end of the line"


def trailing_space(line: str) -> bool:
    if CONFIG.allow_trailing_space:
        return False
    return line.endswith(" ")


def empty_line(line: str) -> bool:
    return line.strip() == ""


def check_whitespace(content: str) -> list[WhiteSpaceError]:
    errors: list[WhiteSpaceError] = []
    empty_line_count = 0
    lines = content.split("\n")
    for i, line in enumerate(lines, start=1):
        if empty_line(line):
            empty_line_count += 1
            if empty_line_count > CONFIG.max_empty_lines:
                errors.append(TooManyEmptyLinesError(i))
            continue

        empty_line_count = 0
        if trailing_space(line):
            errors.append(TrailingSpaceError(i, col=len(line)))

    if CONFIG.require_new_line_eof and not empty_line(lines[-1]):
        errors.append(NoTrailingLineError(len(lines)))

    return errors



import sys

import click
from pyparsing import ParseBaseException
from rich import print

from rapidchecker.whitespace_checks import WhiteSpaceError

from .check import check_format
from .io import get_sys_files, read_sys_file
from .whitespace_checks import check_whitespace


def in_ignore_list(path: str, ignore_list: list[str]) -> bool:
    return any(item in path for item in ignore_list)


def check_file(file_contents: str) -> list[ParseBaseException | WhiteSpaceError]:
    errors: list[ParseBaseException | WhiteSpaceError] = []
    errors.extend(check_format(file_contents))
    errors.extend(check_whitespace(file_contents))
    errors.sort(key=lambda e: e.lineno)
    return errors


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.option("--ignore", multiple=True)
def cli(path: str, ignore: list[str]) -> None:
    found_errors = False

    for filepath in get_sys_files(path):
        if in_ignore_list(str(filepath), ignore):
            print("Skipping", filepath)
            continue
        errors = check_file(read_sys_file(filepath))
        if not errors:
            continue

        found_errors = True
        print(f"[bold]{filepath}[/bold]")
        for error in errors:
            print("\t", str(error))

    if not found_errors:
        print(":heavy_check_mark: ", "No RAPID format errors found!")
    sys.exit(found_errors)



import tomllib
from dataclasses import dataclass, replace
from pathlib import Path


@dataclass
class Config:
    indentation_check: bool = True
    indentation_size: int = 2
    require_uppercase_keywords: bool = True
    max_empty_lines: int = 2
    require_new_line_eof: bool = True
    allow_trailing_space: bool = False

    @classmethod
    def read_config(cls, config_file: Path) -> "Config":
        default_dict = cls()
        if not config_file.exists():
            return default_dict
        with config_file.open("rb") as f:
            user_data = tomllib.load(f)
        return replace(default_dict, **user_data)


CONFIG = Config.read_config(Path("rapidchecker.toml"))
